import json

import pytest


class Helpers:
    def database_config(self, filename: str) -> dict:
        with open(filename, "r") as f:
            data = json.load(f)

        return data


@pytest.fixture
def helpers():
    return Helpers



from unittest.mock import Mock

import pytest

from koia.connector import Database


@pytest.fixture
def mock_mysql_cursor():
    mock_mysql_cursor = Mock()
    mock_mysql_cursor.execute.return_value = None
    mock_mysql_cursor.close.return_value = None
    return mock_mysql_cursor


@pytest.fixture
def mock_mysql_connect(mock_mysql_cursor):
    mock_mysql_connect = Mock()
    mock_mysql_connect.cursor.return_value = mock_mysql_cursor
    mock_mysql_connect.commit.return_value = None
    mock_mysql_connect.close.return_value = mock_mysql_cursor
    return mock_mysql_connect


@pytest.fixture
def mock_mysql(mocker, mock_mysql_connect):
    mock_mysql = mocker.patch("koia.connector.mysql.connector")
    mock_mysql.connect.return_value = mock_mysql_connect
    return mock_mysql


def test_connect(mock_mysql, mock_mysql_connect, mock_mysql_cursor, helpers):
    database_conf = helpers().database_config(
        filename="samples/database_config_sample.json"
    )
    test_connection = Database(config=database_conf)
    assert mock_mysql.connect.called
    assert mock_mysql_connect.cursor.called

    test_connection.reconnect()
    assert mock_mysql_connect.close.called
    assert mock_mysql_cursor.close.called
    assert mock_mysql.connect.call_count == 2
    assert mock_mysql_connect.cursor.call_count == 2

    test_connection.close()
    assert mock_mysql_connect.close.called
    assert mock_mysql_cursor.close.called



import time
from typing import Any, ItemsView, List, Optional

from koia.connector import Database
from koia.QueryBuilder import QueryBuilder


class FetchBuilder(QueryBuilder):
    def __init__(self, database: Database):
        super().__init__(database=database)

    def __order_of_commands(self) -> list:
        """Order of attributes"""
        return [
            "innerJoin",
            "where",
            "orWhere",
            "whereIn",
            "whereNotNull",
            "orderBy",
            "take",
        ]

    def _formalize_sql(self) -> str:
        """Formalize sql query by set attributes from query_conf"""
        sql_query: str = ""
        order_of_commands: list = self.__order_of_commands()
        query_operations: list = []
        query_items: str
        element: str
        key: int
        elem: str

        for query_items in order_of_commands:
            if query_items == "innerJoin":
                query_operations += [
                    element for element in self.query_conf[query_items]
                ]

            else:
                if "where" in query_items:  # Handle where case
                    if len(self.query_conf[query_items]) > 0:
                        query_operations += [
                            element.replace("WHERE", "AND")
                            if len(
                                [elem for elem in query_operations if "WHERE" in elem]
                            )
                            > 0
                            or key > 0
                            else element
                            for key, element in enumerate(self.query_conf[query_items])
                        ]

                if "order" in query_items:
                    if len(self.query_conf[query_items]) > 0:
                        query_operations += [
                            element.replace("ORDER BY", ",")
                            if len(
                                [
                                    elem
                                    for elem in query_operations
                                    if "ORDER BY" in elem
                                ]
                            )
                            > 0
                            or key > 0
                            else element
                            for key, element in enumerate(self.query_conf[query_items])
                        ]

                if "take" in query_items:
                    if len(self.query_conf[query_items]) > 0:
                        query_operations += [
                            element for element in self.query_conf[query_items]
                        ]
                    if len(self.query_conf[query_items]) > 1:
                        raise Exception("You can only once set LIMIT attribute")

        query_operation: str
        for query_operation in query_operations:
            sql_query += f" {query_operation}"

        return sql_query

    def _fetch(self, sql_query) -> Optional[list]:
        """Execute sql query

        Output: List[Any] or None
        Return None if sql query is INSERT OR UPDATE"""
        while True:
            try:
                self.database.cursor.execute(sql_query)
                self.database.connection.commit()
                data = self.database.cursor.fetchall()
                return data
            except Exception as e:
                if str(e) == "2013: Lost connection to MySQL server during query":
                    self.__await_db(error=str(e))
                    self.database.reconnect()
                else:
                    raise Exception(e)

    def __await_db(self, error: str):
        """Sleep for a 3 second"""
        print(error)
        print("Try to reconnect after 3 sec. Zzzz...")
        time.sleep(3)

    def get(self) -> Optional[List[Any]]:
        """Imitize sql SELECT feature"""
        sql_query: str = f"SELECT * FROM {self.query_conf['table']} {self._formalize_sql()}"
        data: Optional[List[Any]] = self._fetch(sql_query=sql_query)
        self.__reload_query_conf()

        return data

    def insert(self, values: dict) -> None:
        """Imitize sql INSERT feature"""
        values_items: ItemsView = values.items()
        sql_columns: str = ",".join(f"`{column[0]}`" for column in values_items)
        sql_values: str = ",".join(str(value[1]) for value in values_items)

        sql_query: str = f"INSERT INTO {self.query_conf['table']} ({sql_columns}) VALUES ({sql_values})"
        self._fetch(sql_query=sql_query)
        self.__reload_query_conf()

    def update(self, values: dict) -> None:
        """Imitize sql UPDATE feature"""
        sql_query: str = self._formalize_sql()
        update_sql_query: str = f"UPDATE {self.query_conf['table']} SET"

        for key, value_item in enumerate(values.items()):
            if key == len(values) - 1:
                update_sql_query += (
                    f""" {value_item[0]}='{value_item[1].replace("'", " ")}'"""
                )
            else:
                update_sql_query += (
                    f""" {value_item[0]}='{value_item[1].replace("'", " ")}',"""
                )
        sql_query = update_sql_query + sql_query

        self._fetch(sql_query=sql_query)
        self.__reload_query_conf()



import mysql.connector


class Database:
    """
    Create MySQLdbConnectoion instance with cursor
    """

    def __init__(self, config: dict = None, autocommit: bool = False) -> None:
        """Initialize connection and cursor"""
        self.config = config
        self.automcommit = autocommit
        self.connection = self._connect(autocommit=self.automcommit)
        self.cursor = self._cursor()

    def _connect(self, autocommit: bool):
        """Connect to db"""
        _connection = mysql.connector.connect(**self.config)
        _connection.autocommit = autocommit
        return _connection

    def _cursor(self):
        """Create cursor"""
        _cursor = self.connection.cursor()
        return _cursor

    def reconnect(self):
        """Re initialize instance"""
        self.connection.close()
        self.cursor.close()
        self.__init__(config=self.config, autocommit=self.automcommit)

    def close(self):
        """Close instance session"""
        self.connection.close()
        self.cursor.close()
